# -*- coding: utf-8 -*-

import logging
import xml.etree.ElementTree as ET
import re

from snowprofile.io._caaml_parse_utils import _parse_str, _parse_numeric, _parse_additional_data, \
    _parse_list, _parse_numeric_list, _search_gml_id, _parse_lat_lon
from snowprofile import _constants


def read_caaml6_xml(filename):
    """
    Parse a CAAML 6 XML-based IACS Snow Profile document and return the associated SnowProfile object.

    Versions up to 6.0.5 are currently supported.

    This parser correctly deal with XML namespaces. One unique namespace URI have to be used for CAAML tags across
    the entire XML file for the parser to work correctly.

    Most of the data stored in the CAAML format is parsed. However, some specific corner cases are not
    covered by this parser. Specifically, when dealing with profiles with unkown snow depth (not compulsoy in CAAML),
    the total depth may be set to 0 and the layer top/bottom height may be negative values.

    Some data is partially parsed. For instance, missing numeric data could be specified as
    inapplicable (there is no value), missing (the value is not available when the data is written and may not exist),
    template (the value will be available later), unknown(the value is not available to the data writer but exists),
    or withheld (the value cannot be disclosed). All these cases are treated in the same way, using a 
    python value ``None``. Some parameters can be represented by text rather than numeric measurements (e.g. grain size could
    be reported in mm or by categories). To facilitate data processing, when categories are used, the package translates the category value into a numerical value
    refering to the center of the category.

    Finally, CAAML may contain some additional data defined by the user. This cannot be parsed, as the structure is
    unknown. Most of this data could be stored and rewritten in a new file but not all, especially not those
    associated with layers in the different profiles.

    Hence, reading a CAAML file and rewritten it with this package may cause some data loss. Most users, however, will experience no problems in this respect.

    :param filename: File path of the CAAML/XML document to parse
    :type filename: str
    :returns: The associated SnowProfile python object (or None in case of errors)
    :rtype: SnowProfile
    """
    # XML parsing
    try:
        tree = ET.parse(filename)
        root = tree.getroot()
    except ET.ParseError as e:
        logging.error(f"Fail to parse CAAML XML-based document {filename}. Error while parsing XML file: {str(e)}")
        return None
    # Identification of CAAML namespace
    if root.tag == "SnowProfile":  # Case no namespace
        ns = None
        nss = ''
    else:
        r = re.match('{(.*)}SnowProfile$', root.tag)
        if r is None:
            logging.error(f"The root element of {filename} is not a SnowProfile element. "
                          "This is not a valid CAAML file.")
            return None
        else:
            ns = r.group(1)
            nss = '{' + ns + '}'
    logging.debug(f"Parsing {filename}. Found CAAML namespace as {nss}")

    # Parsing part by part
    from snowprofile import SnowProfile
    from snowprofile.classes import Time, Observer, Location, Environment, Weather, SurfaceConditions

    # - Time
    time = Time(
        record_time=_parse_str(root, [f'{nss}timeRef/{nss}recordTime/{nss}TimeInstant/{nss}timePosition',
                                      f'{nss}timeRef/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition']),
        record_period=(
            _parse_str(root, f'{nss}timeRef/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
            _parse_str(root, f'{nss}timeRef/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
        report_time=_parse_str(root, f'{nss}timeRef/{nss}dateTimeReport'),
        last_edition_time=_parse_str(root, f'{nss}timeRef/{nss}dateTimeLastEdit'),
        comment=_parse_str(root, f'{nss}timeRef/{nss}metaData/{nss}comment'),
        additional_data=_parse_additional_data(root.find(f'{nss}timeRef/{nss}customData')))

    # - Observer
    observer = Observer(
        source_id=_search_gml_id(root.find(f'{nss}srcRef/{nss}Operation')),
        source_name=_parse_str(root, f'{nss}srcRef/{nss}Operation/{nss}name'),
        source_comment=_parse_str(root, f'{nss}srcRef/{nss}Operation/{nss}metaData/{nss}comment'),
        source_additional_data = _parse_additional_data(root.find(f'{nss}srcRef/{nss}Operation/{nss}customData')))
    contact_persons_1 = root.findall(f'{nss}srcRef/{nss}Operation/{nss}contactPerson')
    contact_persons_2 = root.findall(f'{nss}srcRef/{nss}Person')
    contact_persons = []
    if contact_persons_1 is not None:
        for p in contact_persons_1:
            contact_persons.append(_parse_contact_person(p, nss=nss, ns=ns))
    if contact_persons_2 is not None:
        for p in contact_persons_2:
            contact_persons.append(_parse_contact_person(p, nss=nss, ns=ns))
    if len(contact_persons) > 0:
        observer.contact_persons = contact_persons

    # - Location
    loc = root.find(f'{nss}locRef')
    lat, lon = _parse_lat_lon(loc.find(f'{nss}pointLocation'))
    location = Location(
        id=_search_gml_id(loc),
        name=_parse_str(root, f'{nss}locRef/{nss}name'),
        point_type=_parse_str(root, f'{nss}locRef/{nss}obsPointSubType'),
        aspect=_parse_numeric(root, f'{nss}locRef/{nss}validAspect/{nss}AspectPosition/{nss}position',
                              attribution_table=_constants.aspects),
        elevation=_parse_numeric(root, f'{nss}locRef/{nss}validElevation/{nss}ElevationPosition/{nss}position'),
        slope=_parse_numeric(root, f'{nss}locRef/{nss}validSlopeAngle/{nss}SlopeAnglePosition/{nss}position'),
        latitude=lat,
        longitude = lon,
        country=_parse_str(root, f'{nss}locRef/{nss}country'),
        region=_parse_str(root, f'{nss}locRef/{nss}region'),
        comment=_parse_str(root, f'{nss}locRef/{nss}metaData/{nss}comment'),
        additional_data=_parse_additional_data(root.find(f'{nss}locRef/{nss}customData')))

    # - Environment
    base = f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}weatherCond'
    environment = Environment(
        solar_mask=_parse_solar_mask(root.find(f'{nss}locRef/{nss}solarMask'), nss=nss),
        solar_mask_method_of_measurement=_parse_str(root, f'{nss}locRef/{nss}solarMask/{nss}solarMaskMetaData/{nss}methodOfMeas'),
        solar_mask_uncertainty=_parse_numeric(root, f'{nss}locRef/{nss}solarMask/{nss}solarMaskMetaData/{nss}uncertaintyOfMeas'),
        solar_mask_quality=_parse_str(root, f'{nss}locRef/{nss}solarMask/{nss}solarMaskMetaData/{nss}qualityOfMeas'),
        solar_mask_comment=_parse_str(root, f'{nss}locRef/{nss}solarMask/{nss}solarMaskMetaData/{nss}comment'),
        solar_mask_additional_data=_parse_additional_data(root.find(f'{nss}locRef/{nss}solarMask/{nss}customData')),
        bed_surface=_parse_str(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}bedSurface'),
        bed_surface_comment=_parse_str(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}bedSurfaceComment'),
        litter_thickness=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}litterThickness'),
        ice_thickness=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}iceThickness'),
        low_vegetation_height=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}lowVegetationHeight'),
        LAI=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}lai'),
        forest_presence=_parse_str(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}forestPresence'),
        forest_presence_comment=_parse_str(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}forestComment'),
        sky_view_factor=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}skyViewFactor'),
        tree_height=_parse_numeric(root, f'{nss}locRef/{nss}obsPointEnvironment/{nss}treeHeight'))

    # - Weather
    base = f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}weatherCond'
    weather = Weather(
        cloudiness=_parse_str(root, f'{base}/{nss}skyCond'),
        precipitation=_parse_str(root, f'{base}/{nss}precipTI'),
        air_temperature=_parse_numeric(root, f'{base}/{nss}airTempPres'),
        wind_speed=_parse_numeric(
            root, f'{base}/{nss}windSpd',
            attribution_table=_constants.wind_speed),
        wind_direction=_parse_numeric(
            root, f'{base}/{nss}windDir/{nss}AspectPosition/{nss}position',
            attribution_table=_constants.aspects),
        air_temperature_measurement_height=_parse_numeric(
            root,
            f'{base}/{nss}metaData/{nss}airTempMeasurementHeight'),
        wind_measurement_height=_parse_numeric(
            root,
            f'{base}/{nss}metaData/{nss}windMeasurementHeight'),
        comment = _parse_str(
            root,
            f'{base}/{nss}metaData/{nss}comment'),
        additional_data=_parse_additional_data(root.find(
            f'{base}/{nss}customData')))

    # - Surface Conditions
    base = f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}surfCond'
    surface_conditions = SurfaceConditions(
        surface_roughness=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfRoughness'),
        surface_wind_features=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfWindFeatures'),
        surface_melt_rain_features=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfMeltRainFeatures'),
        surface_features_amplitude=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validAmplitude/{nss}AmplitudePosition/{nss}position',
            factor=0.01),  # cm -> m
        surface_features_amplitude_min=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validAmplitude/{nss}AmplitudeRange/{nss}beginPosition',
            factor=0.01),  # cm -> m
        surface_features_amplitude_max=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validAmplitude/{nss}AmplitudeRange/{nss}endPosition',
            factor=0.01),  # cm -> m
        surface_features_wavelength=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validWavelength/{nss}WavelengthPosition/{nss}position',
            factor=1),  # m
        surface_features_wavelength_min=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validWavelength/{nss}WavelengthRange/{nss}beginPosition',
            factor=1),  # m
        surface_features_wavelength_max=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validWavelength/{nss}WavelengthRange/{nss}endPosition',
            factor=1),  # m
        surface_features_aspect=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}validAspect/{nss}AspectPosition/{nss}position'),
        penetration_ram=_parse_numeric(
            root,
            f'{base}/{nss}penetrationRam',
            factor=0.01),  # cm -> m
        penetration_foot=_parse_numeric(
            root,
            f'{base}/{nss}penetrationFoot',
            factor=0.01),  # cm -> m
        penetration_ski=_parse_numeric(
            root,
            f'{base}/{nss}penetrationSki',
            factor=0.01),  # cm -> m
        lap_presence=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}lapPresence'),
        surface_temperature=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfTemp/{nss}data'),
        surface_temperature_measurement_method=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfTemp/{nss}methodOfMeas'),
        surface_albedo=_parse_numeric(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfAlbedo/{nss}albedo/{nss}albedoMeasurement'),
        surface_albedo_comment=_parse_str(
            root,
            f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfAlbedo/{nss}albedo/{nss}metaData/{nss}comment'),
        spectral_albedo=_parse_spectral_albedo(
            root.find(
                f'{base}/{nss}surfFeatures/{nss}Components/{nss}surfAlbedo/{nss}spectralAlbedo'),
            nss=nss),
        comment=_parse_str(
            root,
            f'{base}/{nss}metaData/{nss}comment'),
        additional_data=_parse_additional_data(root.find(
            f'{base}/{nss}customData')))

    # Creating SnowProfile object
    base = f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}snowPackCond'

    # Profile depth is not taken by default from the profileDepth element, to be coherent
    # with NiViz.
    profile_depth = _parse_numeric(
        root,
        f'{base}/{nss}hS/{nss}Components/{nss}height',
        factor=0.01)  # cm -> m
    if profile_depth is None:
        profile_depth = _parse_numeric(
            root,
            f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}profileDepth',
            factor=0.01)

    sp = SnowProfile(
        id=_search_gml_id(root),
        comment=_parse_str(root, f'{nss}metaData/{nss}comment'),
        profile_comment=_parse_str(
            root,
            f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}metaData/{nss}comment'),
        time=time,
        observer=observer,
        location=location,
        environment=environment,
        weather=weather,
        surface_conditions=surface_conditions,
        application=_parse_str(root, f'{nss}application'),
        application_version=_parse_str(root, f'{nss}applicationVersion'),
        profile_depth=profile_depth,
        profile_depth_std=_parse_numeric(
            root,
            f'{base}/{nss}hSVariability/{nss}Components/{nss}height',
            factor=0.01),
        profile_swe=_parse_numeric(
            root,
            f'{base}/{nss}hS/{nss}Components/{nss}waterEquivalent',
            factor=1),  # kg/m2 is the only one accepted
        profile_swe_std=_parse_numeric(
            root,
            f'{base}/{nss}hSVariability/{nss}Components/{nss}waterEquivalent',
            factor=1),
        new_snow_24_depth=_parse_numeric(
            root,
            f'{base}/{nss}hN24/{nss}Components/{nss}height',
            factor=0.01),  # cm -> m
        new_snow_24_depth_std=_parse_numeric(
            root,
            f'{base}/{nss}hIN/{nss}Components/{nss}height',
            factor=0.01),
        new_snow_24_swe=_parse_numeric(
            root,
            f'{base}/{nss}hN24/{nss}Components/{nss}waterEquivalent',
            factor=1),
        new_snow_24_swe_std=_parse_numeric(
            root,
            f'{base}/{nss}hIN/{nss}Components/{nss}waterEquivalent',
            factor=1),
        snow_transport=_parse_str(
            root,
            f'{base}/{nss}snowTransport'),
        snow_transport_occurence_24=_parse_numeric(
            root,
            f'{base}/{nss}snowTransportOccurrence24'),
        stratigraphy_profile = _parse_stratigraphy(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}stratProfile'),
            nss=nss, profile_depth=profile_depth),
        temperature_profiles = _parse_temperature_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}tempProfile'),
            nss=nss, profile_depth=profile_depth),
        density_profiles = _parse_density_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}densityProfile'),
            nss=nss, profile_depth=profile_depth),
        lwc_profiles = _parse_lwc_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}lwcProfile'),
            nss=nss, profile_depth=profile_depth),
        ssa_profiles = _parse_ssa_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}specSurfAreaProfile'),
            nss=nss, profile_depth=profile_depth),
        hardness_profiles = _parse_hardness_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}hardnessProfile'),
            nss=nss, profile_depth=profile_depth),
        strength_profiles = _parse_strength_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}strengthProfile'),
            nss=nss, profile_depth=profile_depth),
        impurity_profiles = _parse_impurity_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}impurityProfile'),
            nss=nss, profile_depth=profile_depth),
        other_scalar_profiles = _parse_other_scalar_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}otherScalarProfile'),
            nss=nss, profile_depth=profile_depth),
        other_vectorial_profiles = _parse_other_vectorial_profiles(
            root.findall(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}otherVectorialProfile'),
            nss=nss, profile_depth=profile_depth),
        stability_tests = _parse_stability_tests(
            root.find(f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}stbTests'),
            nss=nss, profile_depth=profile_depth),
        additional_data=_parse_additional_data(root.find(
            f'{nss}customData')),
        profile_additional_data = _parse_additional_data(root.find(
            f'{nss}snowProfileResultsOf/{nss}SnowProfileMeasurements/{nss}customData')))

    return sp


def _parse_contact_person(p, nss='', ns=None):
    if p is None:
        return None
    from snowprofile.classes import Person
    return Person(
        id=_search_gml_id(p),
        name=_parse_str(p, f'{nss}name'),
        comment=_parse_str(p, f'{nss}metaData/{nss}comment'),
        additional_data=_parse_additional_data(p.find(f'{nss}customData')))


def _parse_solar_mask(sm_element, nss=''):
    if sm_element is None:
        return None

    from snowprofile.classes import SolarMask

    data = _parse_generic_profile(
        sm_element.findall(f'{nss}Data'),
        {'azimuth': {'path': f'{nss}azimuth', 'type': 'numeric'},
         'elevation': {'path': f'{nss}elevation', 'type': 'numeric'}},
        nss=nss)
    sm = SolarMask(data=data)
    return sm


def _parse_spectral_albedo(sa_element, nss=''):
    if sa_element is None:
        return None

    from snowprofile.classes import SpectralAlbedo

    e = sa_element
    data = _parse_generic_profile(
        e.findall(f'{nss}spectralAlbedoMeasurement'),
        {'min_wavelength': {'path': f'{nss}minWaveLength', 'type': 'numeric'},
         'max_wavelength': {'path': f'{nss}maxWaveLength', 'type': 'numeric'},
         'albedo': {'path': f'{nss}albedo', 'type': 'numeric'},
         'uncertainty': {'path': f'{nss}albedo', 'type': 'numeric', 'attribute': '{nss}uncertainty'},
         'quality': {'path': f'{nss}albedo', 'type': 'str', 'attribute': '{nss}quality'},
         },
        nss=nss)
    sm = SpectralAlbedo(
        data=data,
        comment=_parse_str(e, f'{nss}metaData/{nss}comment'))
    return sm


def _parse_generic_profile(elements, definitions, nss='', min_columns=[]):
    # Eventully get the height to invert depth and height !!
    if elements is None or len(elements) == 0:
        return None
    results = {}
    for key in definitions:
        results[key] = []

    for e in elements:
        for key, value in definitions.items():
            factor = value['numeric_factor'] if 'numeric_factor' in value else 1
            attribute = value['attribute'] if 'attribute' in value else None
            attribution_table = value['attribution_table'] if 'attribution_table' in value else None

            if 'type' in value and value['type'] == 'numeric':
                r = _parse_numeric(e, value['path'], factor=factor, attribute=attribute,
                                   attribution_table=attribution_table)
                if 'adapt_total_depth' in value:
                    r = value['adapt_total_depth'] - r
            elif 'type' in value and value['type'] == 'numeric_list':
                r = _parse_numeric_list(e, value['path'], factor=factor, attribute=attribute)
            else:
                r = _parse_str(e, value['path'], attribute=attribute,
                               attribution_table=attribution_table)
            results[key].append(r)

    # Get rid of columns full of None
    results = {key: value for key, value in results.items() if key in min_columns or set(value) != set([None])}

    return results


def _parse_stratigraphy(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return None

    # Metadata key
    mdk = f'{nss}stratMetaData'

    r = []

    for elem in elements:
        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'grain_1': {'path': f'{nss}grainFormPrimary', 'type': 'str'},
             'grain_2': {'path': f'{nss}grainFormSecondary', 'type': 'str'},
             'grain_size': {'path': f'{nss}grainSize/{nss}Components/{nss}avg', 'type': 'numeric',
                            'numeric_factor': 0.001,  # mm -> m
                            'attribution_table': _constants.grain_sizes},
             'grain_size_max': {'path': f'{nss}grainSize/{nss}Components/{nss}avgMax', 'type': 'numeric',
                                'numeric_factor': 0.001,
                                'attribution_table': _constants.grain_sizes},
             'hardness': {'path': f'{nss}hardness', 'type': 'str'},
             'wetness': {'path': f'{nss}wetness', 'type': 'str'},
             'loc': {'path': f'{nss}layerOfConcern', 'type': 'str'},
             'comment': {'path': f'{nss}metaData/{nss}comment', 'type': 'str'},
             'formation_time': {'path': f'{nss}validFormationTime/{nss}TimeInstant/{nss}timePosition', 'type': 'str'},
             'formation_period_begin': {'path': f'{nss}validFormationTime/{nss}TimePeriod/{nss}beginPosition',
                                        'type': 'str'},
             'formation_period_end': {'path': f'{nss}validFormationTime/{nss}TimePeriod/{nss}endPosition',
                                      'type': 'str'}},
            min_columns=['grain_1', 'grain_size', 'hardness', 'grain_2', 'wetness'],
            nss=nss)

        from snowprofile.profiles import Stratigraphy
        s = Stratigraphy(
            id=_search_gml_id(elem),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r[0] if len(r) > 0 else None


def _parse_temperature_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}tempMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Obs'),
            {'height': {'path': f'{nss}depth', 'type': 'numeric',
                        'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'temperature': {'path': f'{nss}snowTemp', 'type': 'numeric'},
             'uncertainty': {'path': f'{nss}uncertaintyOfMeas', 'type': 'str'},
             'quality': {'path': f'{nss}qualityOfMeas', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import TemperatureProfile
        s = TemperatureProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{nss}tempMetaData/{nss}uncertaintyOfMeas'),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_density_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}densityMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'density': {'path': f'{nss}density', 'type': 'numeric'},
             'uncertainty': {'path': f'{nss}density', 'attribute': 'uncertainty', 'type': 'str'},
             'quality': {'path': f'{nss}density', 'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import DensityProfile
        s = DensityProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            probed_volume = _parse_numeric(elem, f'{mdk}/{nss}probeVolume', factor=1e-6),
            probed_diameter = _parse_numeric(elem, f'{mdk}/{nss}probeDiameter', factor=0.01),
            probed_length = _parse_numeric(elem, f'{mdk}/{nss}probeLength', factor=0.01),
            probed_thickness = _parse_numeric(elem, f'{mdk}/{nss}probedThickness', factor=0.01),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_lwc_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}lwcMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'lwc': {'path': f'{nss}lwc', 'type': 'numeric'},
             'uncertainty': {'path': f'{nss}lwc', 'attribute': 'uncertainty', 'type': 'str'},
             'quality': {'path': f'{nss}lwc', 'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import LWCProfile
        s = LWCProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            probed_thickness = _parse_numeric(elem, f'{mdk}/{nss}probedThickness', factor=0.01),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_ssa_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}specSurfAreaMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        metadata = dict(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            probed_thickness = _parse_numeric(elem, f'{mdk}/{nss}probedThickness', factor=0.01),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')))

        from snowprofile.profiles import SSAProfile, SSAPointProfile

        # Layer profile of SSA
        data1 = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'ssa': {'path': f'{nss}specSurfArea', 'type': 'numeric'},
             'uncertainty': {'path': f'{nss}specSurfArea', 'attribute': 'uncertainty', 'type': 'str'},
             'quality': {'path': f'{nss}specSurfArea', 'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        if data1 is not None and 'top_height' in data1 and len(data1['top_height']) > 0:
            s = SSAProfile(
                **metadata,
                data = data1)
            r.append(s)

        # Point profile of SSA
        height = []
        ssa = []
        e = elem.find(f'{nss}Measurements/{nss}tupleList')
        if e is not None:
            t = e.text
            try:
                for x in t.split(' '):
                    if len(x.strip()) == 0:
                        continue
                    _tuple = x.split(',')
                    _height = _profile_depth - float(_tuple[0]) / 100
                    _ssa = float(_tuple[1])
                    height.append(_height)
                    ssa.append(_ssa)
            except IndexError:
                logging.error('Could not parse the SSA Profile. tupleList is not well formatted '
                              '(not two element at least in the tuple List (depth adn SSA).'
                              f' Value: {_tuple}')
            except ValueError:
                logging.error('Could not parse the SSA Profile. tupleList is not well formatted '
                              '(element of the tuple list that is not recognized as a float value).'
                              f' Value: {_tuple}')
        if len(height) > 0:
            s = SSAPointProfile(
                **metadata,
                data = {'height': height, 'ssa': ssa})
            r.append(s)

    return r


def _parse_hardness_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}hardnessMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        metadata = dict(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            surface_of_indentation = _parse_numeric(elem, f'{mdk}/{nss}surfOfIndentation', factor=0.0001),
            penetration_speed = _parse_numeric(elem, f'{mdk}/{nss}penetrationSpeed', factor=1),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')))

        from snowprofile.profiles import HardnessProfile, HardnessPointProfile

        # Ramsonde Profile of Hardness
        data1 = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'hardness': {'path': f'{nss}hardness', 'type': 'numeric'},
             'weight_hammer': {'path': f'{nss}weightHammer', 'type': 'numeric'},
             'weight_tube': {'path': f'{nss}weightTube', 'type': 'numeric'},
             'n_drops': {'path': f'{nss}nDrops', 'type': 'numeric'},
             'drop_height': {'path': f'{nss}dropHeight', 'type': 'numeric',
                             'numeric_factor': 0.01}},  # cm -> m
            nss=nss)

        if data1 is not None and 'top_height' in data1 and len(data1['top_height']) > 0:
            s = HardnessProfile(
                **metadata,
                data = data1)
            r.append(s)

        # Point profile of Hardness
        height = []
        res = []
        e = elem.find(f'{nss}Measurements/{nss}tupleList')
        if e is not None:
            t = e.text
            try:
                for x in t.split(' '):
                    if len(x.strip()) == 0:
                        continue
                    _tuple = x.split(',')
                    _height = _profile_depth - float(_tuple[0]) / 100
                    _res = float(_tuple[1])
                    height.append(_height)
                    res.append(_res)
            except IndexError:
                logging.error('Could not parse the Hardness Profile. tupleList is not well formatted '
                              '(not two element at least in the tuple List (depth adn SSA).'
                              f' Value: {_tuple}')
            except ValueError:
                logging.error('Could not parse the Hardness Profile. tupleList is not well formatted '
                              '(element of the tuple list that is not recognized as a float value).'
                              f' Value: {_tuple}')
        if len(height) > 0:
            s = HardnessPointProfile(
                **metadata,
                data = {'height': height, 'hardness': res})
            r.append(s)

    return r


def _parse_strength_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}strengthMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'strength': {'path': f'{nss}strengthValue', 'type': 'numeric'},
             'fracture_character': {'path': f'{nss}fractureCharacter', 'type': 'str'},
             'uncertainty': {'path': f'{nss}strengthValue', 'attribute': 'uncertainty', 'type': 'str'},
             'quality': {'path': f'{nss}strengthValue', 'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import StrengthProfile
        s = StrengthProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            probed_area = _parse_numeric(elem, f'{mdk}/{nss}probedArea', factor=1e-4),
            strength_type = _parse_str(elem, f'{mdk}/{nss}strengthType'),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_impurity_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}impurityMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'mass_fraction': {'path': f'{nss}massFraction', 'type': 'numeric'},
             'volume_fraction': {'path': f'{nss}volumeFraction', 'type': 'numeric'},
             'uncertainty': {'path': [f'{nss}volumeFraction', f'{nss}massFraction'],
                             'attribute': 'uncertainty', 'type': 'str'},
             'quality': {'path': [f'{nss}volumeFraction', f'{nss}massFraction'],
                         'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import ImpurityProfile
        s = ImpurityProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            impurity_type = _parse_str(elem, f'{mdk}/{nss}impurity'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            probed_volume = _parse_numeric(elem, f'{mdk}/{nss}probedVolume', factor=1e-6),
            probed_diameter = _parse_numeric(elem, f'{mdk}/{nss}probedDiameter', factor=0.01),
            probed_length = _parse_numeric(elem, f'{mdk}/{nss}probedLength', factor=0.01),
            probed_thickness = _parse_numeric(elem, f'{mdk}/{nss}probedThickness', factor=0.01),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_other_scalar_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}otherScalarMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'data': {'path': f'{nss}value', 'type': 'numeric'},
             'uncertainty': {'path': f'{nss}value', 'attribute': 'uncertainty',
                             'type': 'str'},
             'quality': {'path': f'{nss}value', 'attribute': 'quality', 'type': 'str'}},
            nss=nss)

        from snowprofile.profiles import ScalarProfile
        s = ScalarProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            parameter = _parse_str(elem, f'{mdk}/{nss}parameter'),
            unit = _parse_str(elem, f'{mdk}/{nss}uom'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_other_vectorial_profiles(elements, nss='', profile_depth=0):
    if elements is None or len(elements) == 0:
        return []

    # Metadata key
    mdk = f'{nss}otherVectorialMetaData'

    r = []

    for elem in elements:
        # Metadata

        # Get the profile depth
        profile_depth_local = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}height',
                                             factor=0.01)  # cm -> m
        if profile_depth_local is not None:
            _profile_depth = profile_depth_local
        else:
            _profile_depth = profile_depth if profile_depth is not None else 0

        data = _parse_generic_profile(
            elem.findall(f'{nss}Layer'),
            {'top_height': {'path': f'{nss}depthTop', 'type': 'numeric',
                            'numeric_factor': 0.01, 'adapt_total_depth': _profile_depth},
             'thickness': {'path': f'{nss}thickness', 'type': 'numeric',
                           'numeric_factor': 0.01},  # cm -> m
             'data': {'path': f'{nss}value', 'type': 'numeric_list'},
             'uncertainty': {'path': f'{nss}value', 'attribute': 'uncertainty',
                             'type': 'str'},
             'quality': {'path': f'{nss}value', 'attribute': 'quality', 'type': 'str'}},
            nss=nss, min_columns=['top_height', 'data'])

        from snowprofile.profiles import VectorialProfile
        s = VectorialProfile(
            id=_search_gml_id(elem),
            profile_nr = _parse_numeric(elem, path=f'{nss}profileNr'),
            name = _parse_str(elem, path='.', attribute='name'),
            related_profiles = _parse_list(elem, '.', attribute='relatedProfiles'),
            comment = _parse_str(elem, f'{mdk}/{nss}comment'),
            parameter = _parse_str(elem, f'{mdk}/{nss}parameter'),
            unit = _parse_str(elem, f'{mdk}/{nss}uom'),
            rank = _parse_numeric(elem, f'{mdk}/{nss}rank'),
            method_of_measurement = _parse_str(elem, f'{mdk}/{nss}methodOfMeas'),
            quality_of_measurement = _parse_str(elem, f'{mdk}/{nss}qualityOfMeas'),
            uncertainty_of_measurement = _parse_numeric(elem, f'{mdk}/{nss}uncertaintyOfMeas'),
            record_time = _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimeInstant/{nss}timePosition'),
            record_period = (
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}beginPosition'),
                _parse_str(elem, f'{mdk}/{nss}recordTime/{nss}TimePeriod/{nss}endPosition')),
            profile_depth = profile_depth_local,
            profile_swe = _parse_numeric(elem, f'{mdk}/{nss}hS/{nss}Components/{nss}waterEquivalent'),
            additional_data = _parse_additional_data(elem.find(f'{nss}customData')),
            data = data)
        r.append(s)

    return r


def _parse_stability_tests(element, nss='', profile_depth=0):
    if element is None:
        return []

    r = []

    # RB tests
    for elementtest in element.findall(f'{nss}RBlockTest'):
        from snowprofile.stability_tests import RBStabilityTest, RBStabilityTestResult
        _results = []
        for e in elementtest.findall(f'{nss}failedOn'):
            test_score = _parse_str(e, f'{nss}Results/{nss}testScore')
            m = re.match('RB([0-7])', test_score)
            if m is None:
                logging.error(f'Could not parse Rutschblock score {test_score}.')
                continue
            test_score = int(m.group(1))
            _s = RBStabilityTestResult(
                **_parse_generic_stability_test_result_fields(e, nss=nss, profile_depth=profile_depth),
                test_score = test_score,
                release_type = _parse_str(e, f'{nss}Results/{nss}releaseType'),
                fracture_character = _parse_str(e, f'{nss}Results/{nss}fractureCharacter'))
            _results.append(_s)
        s = RBStabilityTest(
            **_parse_generic_stability_test_fields(elementtest, nss=nss, profile_depth=profile_depth),
            results = _results)
        r.append(s)

    # CT tests
    for elementtest in element.findall(f'{nss}ComprTest'):
        from snowprofile.stability_tests import CTStabilityTest, CTStabilityTestResult
        _results = []
        for e in elementtest.findall(f'{nss}failedOn'):
            test_score = _parse_str(e, f'{nss}Results/{nss}testScore', attribution_table=_constants.CT_scores)
            try:
                test_score = int(test_score)
            except ValueError:
                logging.error(f'Could not parse ECT score {test_score}.')
                continue
            _s = CTStabilityTestResult(
                **_parse_generic_stability_test_result_fields(e, nss=nss, profile_depth=profile_depth),
                test_score = test_score,
                fracture_character = _parse_str(e, f'{nss}Results/{nss}fractureCharacter'))
            _results.append(_s)
        s = CTStabilityTest(
            **_parse_generic_stability_test_fields(elementtest, nss=nss, profile_depth=profile_depth),
            results = _results)
        r.append(s)

    # ECT tests
    for elementtest in element.findall(f'{nss}ExtColumnTest'):
        from snowprofile.stability_tests import ECTStabilityTest, ECTStabilityTestResult
        _results = []
        for e in elementtest.findall(f'{nss}failedOn'):
            test_score = _parse_str(e, f'{nss}Results/{nss}testScore')
            if test_score == 'ECTPV':
                test_score = 0
                propagation = True
            else:
                m = re.match('ECT([NP])([0-9]*)', test_score)
                if m is None:
                    logging.error(f'Could not parse Rutschblock score {test_score}.')
                    continue
                test_score = int(m.group(2))
                propagation = m.group(1) == 'P'
            _s = ECTStabilityTestResult(
                **_parse_generic_stability_test_result_fields(e, nss=nss, profile_depth=profile_depth),
                test_score = test_score,
                propagation = propagation)
            _results.append(_s)
        s = ECTStabilityTest(
            **_parse_generic_stability_test_fields(elementtest, nss=nss, profile_depth=profile_depth),
            results = _results)
        r.append(s)

    # PST tests
    for elementtest in element.findall(f'{nss}PropSawTest'):
        from snowprofile.stability_tests import PSTStabilityTest
        e = elementtest.find(f'{nss}failedOn')
        if e is not None:
            s = PSTStabilityTest(
                **_parse_generic_stability_test_fields(elementtest, nss=nss, profile_depth=profile_depth),
                **_parse_generic_stability_test_result_fields(e, nss=nss, profile_depth=profile_depth),
                column_length = _parse_numeric(e, f'{nss}Results/{nss}columnLength', factor=0.01),
                cut_length = _parse_numeric(e, f'{nss}Results/{nss}cutLength', factor=0.01),
                propagation = _parse_str(e, f'{nss}Results/{nss}fracturePropagation'))
            r.append(s)

    # Shear frame tests
    for elementtest in element.findall(f'{nss}ShearFrameTest'):
        from snowprofile.stability_tests import ShearFrameStabilityTest, ShearFrameStabilityTestResult
        _results = []
        for e in elementtest.findall(f'{nss}failedOn'):
            _s = ShearFrameStabilityTestResult(
                **_parse_generic_stability_test_result_fields(e, nss=nss, profile_depth=profile_depth),
                force = _parse_numeric(e, f'{nss}Results/{nss}failureForce'),
                fracture_character = _parse_str(e, f'{nss}Results/{nss}fractureCharacter'))
            _results.append(_s)
        s = ShearFrameStabilityTest(
            **_parse_generic_stability_test_fields(elementtest, nss=nss, profile_depth=profile_depth),
            results = _results)
        r.append(s)

    return r


def _parse_generic_stability_test_fields(elem, nss='', profile_depth=0):
    return dict(
        id=_search_gml_id(elem),
        name = _parse_str(elem, path='.', attribute='name'),
        test_nr = _parse_numeric(elem, path=f'{nss}testNr'),
        comment = _parse_str(elem, f'{nss}metaData/{nss}comment'),
        additional_data = _parse_additional_data(elem.find(f'{nss}customData')))


def _parse_generic_stability_test_result_fields(elem, nss='', profile_depth=0):
    depth = _parse_numeric(elem, f'{nss}Layer/{nss}depthTop', factor = 0.01)
    return dict(
        height = profile_depth - depth if depth is not None else None,
        layer_thickness = _parse_numeric(elem, f'{nss}Layer/{nss}thickness', factor = 0.01),
        grain_1 = _parse_str(elem, f'{nss}Layer/{nss}grainFormPrimary'),
        grain_2 = _parse_str(elem, f'{nss}Layer/{nss}grainFormSecondary'),
        grain_size = _parse_numeric(elem, f'{nss}Layer/{nss}grainSize/{nss}Components/{nss}avg', factor=0.001,
                                    attribution_table=_constants.grain_sizes),
        grain_size_max = _parse_numeric(elem, f'{nss}Layer/{nss}grainSize/{nss}Components/{nss}avgMax', factor=0.001,
                                        attribution_table=_constants.grain_sizes),
        layer_formation_time = _parse_str(elem,
                                          f'{nss}Layer/{nss}validFormationTime/{nss}TimeInstant/{nss}timePosition'),
        layer_formation_period = (
            _parse_str(elem, f'{nss}Layer/{nss}validFormationTime/{nss}TimePeriod/{nss}beginPosition'),
            _parse_str(elem, f'{nss}Layer/{nss}validFormationTime/{nss}TimePeriod/{nss}endPosition')),
        layer_comment = _parse_str(elem, f'{nss}Layer/{nss}metaData/{nss}comment'),
        layer_additional_data = _parse_additional_data(elem.find(f'{nss}Layer/{nss}customData')))



# -*- coding: utf-8 -*-

import re
import logging
import xml.etree.ElementTree as ET

import pyproj


def _parse_str(root, path, clean=True, attribute=None, attribution_table=None):
    """
    Search for an element described by path in the XML Element root
    and parse its content as a string.

    :param root: A XML Element (or None)
    :param path: Path to the searched data (str or list, in case this is a list, take the first non-void element)
    :param clean: Apply strip() on the resulting string
    :param attribute: Parse the content of an attribute of the given element rather than the text content.
    :returns: Parsed string or None if not found
    """
    if root is None:
        return None

    if isinstance(path, list):
        for p in path:
            f = root.find(p)
            if f is not None:
                break
    else:
        f = root.find(path)

    if f is not None:
        if attribute is None:
            r = f.text
        else:
            if attribute in f.attrib:
                r = f.attrib[attribute]
            else:
                return None
        if clean and r is not None:
            r = r.strip()
        if attribution_table is not None and r in attribution_table:
            return attribution_table[r]

        if r in ["inapplicable", "missing", "template", "unknown", "withheld"]:
            return None
        return r


def _parse_numeric(root, path, factor=1, attribute=None, attribution_table=None):
    """
    Search for an element described by path in the XML Element root
    and parse its content as a floating-point number.

    :param root: A XML Element (or None)
    :param path: Path to the searched data
    :param factor: A factor to apply to the parsed float
    :param attribute: Parse the content of an attribute of the given element rather than the text content.
    :returns: Parsed float or None if not found
    """
    f = _parse_str(root, path, attribute=attribute, attribution_table=attribution_table)

    if f is not None:
        try:
            f = float(f)
            return f * factor
        except Exception:
            return None


def _parse_numeric_list(root, path, factor=1, attribute=None):
    """
    Search for an element described by path in the XML Element root
    and parse its content as a list of floating-point number.

    :param root: A XML Element (or None)
    :param path: Path to the searched data
    :param factor: A factor to apply to the parsed float
    :param attribute: Parse the content of an attribute of the given element rather than the text content.
    :returns: Parsed list of float or None if not found
    """
    if root is None:
        return None

    if isinstance(path, list):
        for p in path:
            f = root.find(p)
            if f is not None:
                break
    else:
        f = root.find(path)

    if f is not None:
        if attribute is None:
            r = f.text.strip()
        else:
            if attribute in f.attrib:
                r = f.attrib[attribute]
            else:
                return None
        try:
            fl = r.split()
            rl = [float(x) * factor for x in fl]
            return rl
        except Exception:
            return None


def _search_gml_id(element):
    if element is None:
        return None
    for key, value in element.attrib.items():
        r = re.match('{.*}id$', key)
        if r is not None or key == 'id':
            return value


def _parse_lat_lon(pointlocation):
    """
    Return a tuple latitude, longitude

    NB: this is written the other side in the file as CRS84 is used.
    """
    if pointlocation is None:
        return None, None
    for elem in pointlocation:
        r = re.match('{.*}Point$', elem.tag)
        if r is not None or elem.tag == 'Point':
            srs = elem.attrib['srsName'] if 'srsName' in elem.attrib else None
            if srs is None:
                logging.warning('Could not determine SRS. We imagine CRS84 and will parse lon/lat accordingly.')
            for e in elem:
                r = re.match('{.*}pos$', e.tag)
                if r is not None or e.tag == 'pos':
                    try:
                        sp = e.text.strip().split()
                        if srs is not None:
                            t = pyproj.Transformer.from_crs(srs, 'EPSG:4326')
                            return t.transform(*[float(x) for x in sp])
                        else:  # Suppose CRS84 : lon/lat
                            return float(sp[1]), float(sp[0])
                    except Exception:
                        logging.warning('Could not parse latitude/longitude')

    logging.error('Could not parse latitude/longitude.')
    return None, None


def _parse_list(*args, **kwargs):
    r = _parse_str(*args, **kwargs)
    if r is None:
        return []
    else:
        return r.split()


def _parse_additional_data(ad_element, origin='caamlxml6'):
    if ad_element is not None:
        data = ''
        for e in ad_element:
            data += ET.tostring(e, encoding='utf-8').decode('utf-8')
        return {'data': data, 'origin': origin}
    return None



# -*- coding: utf-8 -*-

import logging
import datetime

from snowprofile import _utils


def read_mf_bdclim(numposte, date, db_config={}):
    """
    Read snow profile from a database as stored in the Meteo-France climatological database.
    This routine is designed for internal use at Meteo-France only.

    You have to provide database url and creedentials:

    - Via the ``bd_config`` key
    - Or in the ``[io_mf_bdclim]`` key of the snowprofile configuration file (see :ref:`configuration`)

    Keys to configure connexion are ``host``, ``port``, ``dbname``, ``user``, ``password``.

    :param numposte: numposte
    :type numposte: str
    :param date: date to read
    :type date: python datetime object or str
    :param db_config: The information to connect to the database
    :type db_config: dict
    :returns: The SnowProfile object containin the data
    :rtype: SnowProfile
    """
    date = _utils.check_date(date)
    conn = _mf_conn(**db_config)

    # Read database
    metadata_poste = _get_poste_info(conn, numposte)
    metadata = _get_metadata_obs(conn, numposte, date)

    if metadata is None:
        conn.close()
        raise ValueError(f"Could not find data at date {date.isoformat(sep=' ')}")

    profil, profil_ram, profil_t, maxdepth = _get_profil(conn, numposte, date)

    conn.close()

    # Check topdepth
    if metadata['totdepth'] is None:
        metadata['totdepth'] = maxdepth
    else:
        if not maxdepth == metadata['totdepth']:
            logging.error(f'Incompatbile maximum depth: {maxdepth * 100}cm in the database and '
                          f"{metadata['totdepth'] * 100} in the data.")
            metadata['totdepth'] = maxdepth

    # Process data
    from snowprofile import SnowProfile
    from snowprofile.profiles import DensityProfile, Stratigraphy, HardnessProfile, TemperatureProfile, LWCProfile
    from snowprofile.classes import Location, Weather, Time

    loc = Location(
        id=f'numposte{numposte}',
        name=f"{metadata_poste['name']} ({metadata_poste['name_detail']})",
        latitude=metadata_poste['lat'],
        longitude=metadata_poste['lon'],
        elevation=metadata_poste['elevation'],
        slope=metadata_poste['slope'],
        aspect=metadata_poste['aspect'])

    # Stratigraphy
    s = Stratigraphy(data={
        'top_height': [p[0] for p in profil],
        'thickness': [p[1] for p in profil],
        'grain_1': [p[2] for p in profil],
        'grain_2': [p[3] for p in profil],
        'grain_size': [p[4] for p in profil],
        'hardness': [p[5] for p in profil],
        'wetness': [p[6] for p in profil]})

    # Density profile
    d_v = [p[7] for p in profil]
    if len(set(d_v)) > 0 and set(d_v) != set([None, ]):
        d = [DensityProfile(
            method_of_measurement='Snow Cylinder',
            data={
                'top_height': [p[0] for i, p in enumerate(profil) if d_v[i] is not None],
                'thickness': [p[1] for i, p in enumerate(profil) if d_v[i] is not None],
                'density': [p[7] for i, p in enumerate(profil) if d_v[i] is not None]}), ]
    else:
        d = []

    # LWC
    lwc_v = [p[9] for p in profil]
    if len(set(lwc_v)) > 0 and set(lwc_v) != set([None, ]):
        lwc = [LWCProfile(data={
            'top_height': [p[0] for i, p in enumerate(profil) if lwc_v[i] is not None],
            'thickness': [p[1] for i, p in enumerate(profil) if lwc_v[i] is not None],
            'lwc': [p[9] for i, p in enumerate(profil) if lwc_v[i] is not None]}), ]
    else:
        lwc = []

    # TODO: Faire qqch de mesure cisso ?  <27-01-25, Léo Viallon-Galinier> #

    # RAM Profile
    if len(profil_ram) > 0:
        r = [HardnessProfile(data={
            'top_height': [p[0] for p in profil_ram],
            'thickness': [p[1] for p in profil_ram],
            'hardness': [p[2] for p in profil_ram]}), ]
    else:
        r = []

    # Temp profile
    if len(profil_t) > 0:
        t = [TemperatureProfile(data={
            'height': [p[0] for p in profil_t],
            'temperature': [p[1] for p in profil_t]}), ]
    else:
        t = []

    # Observer
    observer = _utils.get_default_observer(key='io_mf_bdclim')

    # Time
    time = Time(record_time=date)

    sp = SnowProfile(
        id=f'numposte{numposte}-{date.strftime("%Y%m%d%H%M")}',
        comment=metadata['comment'],
        profile_depth=metadata['totdepth'],
        profile_swe=metadata['lwc'],
        location=loc,
        time=time,
        weather=Weather(cloudiness=metadata['cloudiness'],
                        precipitation=metadata['precipitation'],
                        air_temperature=float(metadata['t']) if metadata['t'] is not None else None),
        observer=observer,
        stratigraphy_profile=s,
        density_profiles=d,
        hardness_profiles=r,
        temperature_profiles=t,
        snow_transport='Drifting snow' if metadata['snow_transport'] is True else None,
        lwc_profiles=lwc)

    return sp


def search_mf_bdclim_dates(numposte, date_min, date_max=None, db_config={}):
    """
    Get the dates of observation for given num poste and between
    date_min and date_max.

    See

    :param numposte: numposte
    :type numposte: str
    :param date_min: begin date for search
    :type date_min: python datetime object or str
    :param date_max: end date for search
    :type date_max: python datetime object or str
    :param db_config: The information to connect to the database
    :type db_config: dict
    :returns: list of available dates
    :rtype: list of python datetime objects
    """

    date_min = _utils.check_date(date_min)
    if date_max is not None:
        date_max = _utils.check_date(date_max)
    else:
        date_max = datetime.datetime.now()
    sql = ("SELECT DISTINCT dat FROM donnees_profil_neige "
           "WHERE num_poste='{numposte}' AND dat >= '{beg}' AND dat <= '{end}';".format(numposte=numposte,
                                                                                        beg=date_min.isoformat(sep=' '),
                                                                                        end=date_max.isoformat(sep=' '),
                                                                                        )
           )
    conn = _mf_conn(**db_config)
    with conn.cursor() as cur:
        cur.execute(sql)
        dates = cur.fetchall()
    conn.close()

    dates = [d[0] for d in dates]
    return dates


def _get_poste_info(conn, numposte):
    """
    Get the info from poste

    :param conn: connection to the database
    :type conn: psycopg2 connexion
    :param numposte: numposte
    :type numposte: str
    :returns: Information related to poste
    :rtype: dict
    """
    sql = ("SELECT nom_usuel, lieu_dit, alti, lat_dg, lon_dg, exposition_nivo, pente_nivo"
           " FROM poste_nivo WHERE num_poste='{}';".format(numposte))
    with conn.cursor() as cur:
        cur.execute(sql)
        data = cur.fetchone()

    aspect_raw = str(data[5])
    if aspect_raw in _correspStrAspect:
        aspect = _correspStrAspect[aspect_raw]
    else:
        aspect = None

    r = {'name': data[0],
         'name_detail': data[1],
         'elevation': int(data[2]),
         'slope': int(data[6]),
         'aspect': aspect,
         'lat': float(data[3]),
         'lon': float(data[4]),
         }
    return r


def _get_profil(conn, numposte, date):
    """
    Get the standard profile (grain shape, depths, thicknesses) for
    a given numposte and date.

    :param conn: connection to the database
    :type conn: psycopg2 connexion
    :param numposte: numposte
    :type numposte: str
    :param date: Date of the observation
    :type date: python datetime object
    :returns: Profiles on the form of a list of lines. Standard profile, RAM profile and Temperature profile.
              Maximum depth is also returned.
              For std profile: topdepth, thickness, grain primary, grain secondary, grain size,
                             hardness_class, lwc_class, density (kg/m3), hardness (daN), lwc (%)
              For RAM profile: topdepth, thickness, value (daN)
              For T profile: depth, value (Celsius degree)
    :rtype: list, list, list, int
    """
    sql = """SELECT hauteur, resist, t_neige, cod_type_grain1, cod_type_grain2, \
             diam_grain, cod_dur_grain, cod_u_neige, masse_vol, teneur_eau, cisaillt_cal, \
             epaisseur_couche_strati, epaisseur_couche_resist \
             FROM donnees_profil_neige \
             WHERE num_poste=\'{numposte}\' AND dat=\'{date}\' \
             ORDER BY hauteur DESC;
             """.format(numposte=numposte, date=date.isoformat(sep=' '))

    with conn.cursor() as cur:
        cur.execute(sql)
        data = cur.fetchall()

    profil_std = []
    profil_ram = []
    profil_t = []
    maxdepth = 0

    for i, line in enumerate(data):
        # Standard profile
        if line[3] is not None:
            topdepth = int(line[0])
            maxdepth = max(topdepth, maxdepth)
            ep = line[11]
            if line[3] == -8:  # Special case for partial pit
                continue
            g1 = _correspGrainForm[line[3]]
            g2 = _correspGrainForm[line[4]] if line[4] is not None else _correspGrainForm[line[3]]
            diam = float(line[5]) / 1e3 if line[5] is not None else None
            hardness = _correspHardness[line[6]] if line[6] in _correspHardness else None
            lwc = _correspLwc[line[7]] if line[7] in _correspLwc else None
            density = float(line[8]) if line[8] is not None else None
            cisaillt = float(line[10]) if line[10] is not None else None
            lwc_m = float(line[9]) if line[9] is not None else None


            # ep processing
            # sometimes epaisseur_couche_strati is not defined...
            if ep is None:
                ep = topdepth
                ii = i + 1
                for j in range(ii, len(data)):
                    if data[j][3] is not None:
                        ep = topdepth - int(data[j][0])
                        break
            ep = int(ep)

            profil_std.append([topdepth / 100, ep / 100, g1, g2,
                               diam, hardness, lwc,
                               density, cisaillt, lwc_m]
                              )

        # RAM profile
        if line[1] is not None:
            topdepth = int(line[0])
            maxdepth = max(topdepth, maxdepth)
            ep = line[12]
            ram = float(line[1]) * 9.81  # Convert kgf in N

            # ep processing
            # sometimes epaisseur_couche_resist is not defined...
            if ep is None:
                ep = topdepth
                ii = i + 1
                for j in range(ii, len(data)):
                    if data[j][1] is not None:
                        ep = topdepth - int(data[j][0])
                        break
            ep = int(ep)

            profil_ram.append([topdepth / 100, ep / 100, ram])

        # Temperature profile
        if line[2] is not None:
            depth = int(line[0])
            maxdepth = max(depth, maxdepth)
            t = float(line[2])

            profil_t.append([depth / 100, t])

    return profil_std, profil_ram, profil_t, maxdepth / 100


def _get_metadata_obs(conn, numposte, date):
    """
    Get the metadata associated to profile observation

    :param conn: connection to the database
    :type conn: psycopg2 connexion
    :param numposte: numposte
    :type numposte: str
    :param date: Date of the observation
    :type date: python datetime object
    :returns: Information related to obsrvation
    :rtype: dict
    """
    sql = """SELECT t, ww_profil_neige, hauteur_neige, equivalent_eau, comment_court, comment_long \
             from infos_profil_neige WHERE num_poste='{numposte}' and dat='{date}'
             """.format(numposte=numposte, date=date.isoformat(sep=' '))
    with conn.cursor() as cur:
        cur.execute(sql)
        data = cur.fetchone()

    if data is None:
        return None

    comment = "{}\n{}".format(data[4] if data[4] is not None else '',
                              data[5] if data[5] is not None else '')

    ww = _correspWWSkyCond[data[1]] if data[1] is not None and data[1] in _correspWWSkyCond else None
    if ww is not None:
        ww = {'cloudiness': ww[0], 'precipitation': ww[1], 'snow_transport': ww[2]}
    else:
        ww = {'cloudiness': None, 'precipitation': None, 'snow_transport': None}

    r = {'t': data[0],
         'totdepth': int(data[2]) / 100 if data[2] is not None else None,
         'lwc': int(data[3]) if data[3] is not None else None,
         'comment': comment,
         **ww
         }
    return r


def _mf_conn(**kwargs):
    import psycopg2

    # Read from config the database details if not provided in kwargs
    if 'host' not in kwargs:
        config = _utils.get_config()
        if 'io_mf_bdclim' in config:
            c = config['io_mf_bdclim']
            if 'host' in c:
                kwargs['host'] = c['host']
            if 'port' in c:
                kwargs['port'] = c['port']
            if 'user' in c:
                kwargs['user'] = c['user']
            if 'password' in c:
                kwargs['password'] = c['password']
            if 'dbname' in c:
                kwargs['dbname'] = c['dbname']

    # If config not found and nothing provided in kwargs: cannot connect.
    if 'host' not in kwargs:
        logging.critical('io_mf_bdclim: Could not connect to the database. '
                         'Please provide host and creedentials.')
        raise ValueError('Host not known to connect to database.')

    # Connect to the database
    return psycopg2.connect(**kwargs)


_correspStrAspect = {
    "0": 0.0,     # N
    "1": 0.0,     # N
    "2": 22.5,    # NNE
    "3": 22.5,    # NNE
    "4": 45.0,    # NE
    "5": 45.0,    # NE
    "6": 67.5,    # ENE
    "7": 67.5,    # ENE
    "8": 90.0,    # E
    "9": 90.0,    # E
    "10": 90.0,   # E
    "11": 112.5,  # ESE
    "12": 112.5,  # ESE
    "13": 135.0,  # SE
    "14": 135.0,  # SE
    "15": 157.5,  # SSE
    "16": 157.5,  # SSE
    "17": 180.0,  # S
    "18": 180.0,  # S
    "19": 180.0,  # S
    "20": 202.5,  # SSW
    "21": 202.5,  # SSW
    "22": 225.0,  # SW
    "23": 225.0,  # SW
    "24": 225.0,  # SW
    "25": 247.5,  # WSW
    "26": 270.0,  # W
    "27": 270.0,  # W
    "28": 270.0,  # W
    "29": 292.5,  # WNW
    "30": 157.5,  # WNW
    "31": 315.0,  # NW
    "32": 315.0,  # NW
    "33": 337.5,  # NNW
    "34": 337.5,  # NNW
    "35": 0.0,    # N
    "36": 0.0,    # N
    # 96 Crete
    # 97 Fond de vallee
    # 98 Plateau
}

_correspGrainForm = {
    1: 'PP',
    2: 'DF',
    3: 'RG',
    4: 'FC',
    5: 'DH',
    6: 'MF',
    7: 'IF',
    8: 'SH',
    9: 'PPgp',
}

_correspHardness = {
    1: 'F',
    2: '4F',
    3: '1F',
    4: 'P',
    5: 'K',
}

_correspLwc = {
    1: 'D',
    2: 'M',
    3: 'W',
    4: 'V',
    5: 'S',
}

_correspWWSkyCond = {  # cloudiness, precipitation, snow transport
    0: ['CLR', 'Nil', None],  # RAS
    1: ['BKN', 'Nil', None],  # Nuageux à couvert (>4/8)
    2: [None, 'Nil', None],   # Vent
    3: [None, 'Nil', True],   # Chasse-neige (au ras du sol ou en volutes)
    4: ['X', 'Nil', None],    # Brouillard
    6: [None, 'RA', None],    # Pluie
    7: [None, 'SN', None],    # Neige
    9: [None, None, None],    # Orage
    -9: [None, None, None],   # type de temps inconnu
    None: [None, None, None],   # type de temps inconnu
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

import snowprofile


def gen_snowprofile(offset=0):
    sp = snowprofile.SnowProfile()
    sp.location.name = 'Col de Porte'
    sp.location.latitude = 45.295043
    sp.location.longitude = 5.76559
    import datetime
    sp.time.record_time = datetime.datetime(2019, 12, 25, 10, 0)
    sp.time.report_time = datetime.datetime(2023, 11, 2)

    dp = snowprofile.profiles.DensityProfile(
        method_of_measurement="Snow Cutter",
        quality_of_measurement="Good",
        probed_thickness=0.03,  # 3cm cutter thickness
        data = {'top_height': [1.2 + offset, 1.1 + offset, 1.0 + offset, 0.50],
                'thickness': [0.1, 0.1, 0.5, 0.5],
                'density': [75, 100, 180, 230]}
    )
    sp.density_profiles.append(dp)
    return sp


class TestMerge(unittest.TestCase):

    def test_merge(self):
        sp1 = gen_snowprofile()
        sp1_ref = gen_snowprofile()
        sp2 = gen_snowprofile(offset=1)

        sp1.merge(sp2)

        assert sp1.location.name == sp1_ref.location.name
        assert len(sp1.density_profiles) == len(sp1_ref.density_profiles) + len(sp2.density_profiles)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import os.path
import unittest
import tempfile

from lxml import etree

import snowprofile

_here = os.path.dirname(os.path.realpath(__file__))


class TestIOCAAML6XML(unittest.TestCase):

    def test_read_caaml6_xml_default_example(self):
        """
        Test file provided on caaml.org website (SnowProfile_IACS_SLF22950.xml)
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'SnowProfile_IACS_SLF22950.xml'))

        assert sp.observer.source_name == 'WSL Insitute for Snow and Avalanche Research SLF'
        assert sp.location.name == 'Chörbschhorn - Hanengretji - Davos'
        assert sp.application == 'SnowProfiler'
        assert sp.weather.cloudiness == 'FEW'
        assert sp.profile_depth == 1.831
        assert len(sp.temperature_profiles) == 1
        assert len(sp.density_profiles) == 1
        assert len(sp.impurity_profiles) == 1
        assert sp.impurity_profiles[0].impurity_type == 'Black Carbon'
        assert sp.location.latitude - 46.7911050 < 1e-6
        assert sp.location.longitude - 9.7732374 < 1e-6

    def test_read_write_caaml6_xml_example2(self):
        """
        A rather simple profile, but with diffrent data compated to the basic test file provided on caaml.org website.
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'TestProfile2.caaml'))

        with tempfile.TemporaryDirectory(prefix='snowprofiletests') as dirname:
            filename = os.path.join(dirname, 'testcaaml.caaml')
            snowprofile.io.write_caaml6_xml(sp, filename)

            # Test CAAML6 validity
            xmlschema_path = os.path.join(_here, 'resources', 'CAAMLv6.0.5_SnowProfileIACS.xsd')
            xmlschema_doc = etree.parse(xmlschema_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)

            xml_doc = etree.parse(filename)
            result = xmlschema.validate(xml_doc)
            assert result, f'Invalid CAAML output:\n{xmlschema.error_log}'

            # Re-read
            sp_reread = snowprofile.io.read_caaml6_xml(filename)

        assert snowprofile.io.to_json(sp) == snowprofile.io.to_json(sp_reread)

    def test_write_caaml6_xml_example3_v605(self):
        """
        Test Profile 3: a rather complete test snow observation.
        Write in CAAML 6.0.5 version
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'TestProfile3.caaml'))

        with tempfile.TemporaryDirectory(prefix='snowprofiletests') as dirname:
            filename = os.path.join(dirname, 'testcaaml.caaml')
            snowprofile.io.write_caaml6_xml(sp, filename, version='6.0.5')

            # Test CAAML6 validity
            xmlschema_path = os.path.join(_here, 'resources', 'CAAMLv6.0.5_SnowProfileIACS.xsd')
            xmlschema_doc = etree.parse(xmlschema_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)

            xml_doc = etree.parse(filename)
            result = xmlschema.validate(xml_doc)
            assert result, f'Invalid CAAML output:\n{xmlschema.error_log}'

    def test_read_write_caaml6_xml_example3_v606(self):
        """
        Test Profile 3: a rather complete test snow observation.
        Write in CAAML 6.0.6 version and ensure no data is lost.
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'TestProfile3.caaml'))

        with tempfile.TemporaryDirectory(prefix='snowprofiletests') as dirname:
            filename = os.path.join(dirname, 'testcaaml.caaml')
            snowprofile.io.write_caaml6_xml(sp, filename, version='6.0.6')

            # Test CAAML6 validity
            xmlschema_path = os.path.join(_here, 'resources', 'CAAMLv6.0.6_SnowProfileIACS.xsd')
            xmlschema_doc = etree.parse(xmlschema_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)

            xml_doc = etree.parse(filename)
            result = xmlschema.validate(xml_doc)
            assert result, f'Invalid CAAML output:\n{xmlschema.error_log}'

            # Re-read
            sp_reread = snowprofile.io.read_caaml6_xml(filename)

        assert snowprofile.io.to_json(sp) == snowprofile.io.to_json(sp_reread)

    def test_write_caaml6_xml_example4_v605(self):
        """
        Test Profile 4: A profile with exotic things not present in test 3
        Write in CAAML 6.0.5 version
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'TestProfile3.caaml'))

        with tempfile.TemporaryDirectory(prefix='snowprofiletests') as dirname:
            filename = os.path.join(dirname, 'testcaaml.caaml')
            snowprofile.io.write_caaml6_xml(sp, filename, version='6.0.5')

            # Test CAAML6 validity
            xmlschema_path = os.path.join(_here, 'resources', 'CAAMLv6.0.5_SnowProfileIACS.xsd')
            xmlschema_doc = etree.parse(xmlschema_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)

            xml_doc = etree.parse(filename)
            result = xmlschema.validate(xml_doc)
            assert result, f'Invalid CAAML output:\n{xmlschema.error_log}'

    def test_read_write_caaml6_xml_example4_v606(self):
        """
        Test Profile 4: A profile with exotic things not present in test 3
        Write in CAAML 6.0.6 version and ensure no data is lost.
        """
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'TestProfile4.caaml'))

        with tempfile.TemporaryDirectory(prefix='snowprofiletests') as dirname:
            filename = os.path.join(dirname, 'testcaaml.caaml')
            snowprofile.io.write_caaml6_xml(sp, filename, version='6.0.6')

            # Test CAAML6 validity
            xmlschema_path = os.path.join(_here, 'resources', 'CAAMLv6.0.6_SnowProfileIACS.xsd')
            xmlschema_doc = etree.parse(xmlschema_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)

            xml_doc = etree.parse(filename)
            result = xmlschema.validate(xml_doc)
            assert result, f'Invalid CAAML output:\n{xmlschema.error_log}'

            # Re-read
            sp_reread = snowprofile.io.read_caaml6_xml(filename)

        assert snowprofile.io.to_json(sp) == snowprofile.io.to_json(sp_reread)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""
Base Classes and field validators that are used in the snowprofile, classes, profiles
stability_tests packages.
"""

import datetime
import typing
import logging
import re
import sys

import pydantic
import pydantic.json_schema
import pandas as pd
import numpy as np


class AdditionalData(pydantic.BaseModel):
    data: typing.Any
    origin: typing.Optional[str] = None


def force_utc(value: str | datetime.datetime | None) -> typing.Optional[datetime.datetime]:
    """
    Parse to a datetime object and force the tzinfo to be defined in a python datetime object.

    In case the tzinfo is not provided, assume UTC.
    """
    if value is None:
        return None
    if isinstance(value, str):
        if sys.version_info.major == 3 and sys.version_info.minor < 11:
            # On python <= 3.10, datetime.fromisoformat have troubles with decimal seconds as
            # prodived by some other libraries such as NiViz. Get rid of this unnecessary precision.
            m = re.match(r'([0-9]{4}-[0-9]{2}-[0-9]{2}[ T]?[0-9]{2}:[0-9]{2}:[0-9]{2})\.[0-9]*', value)
            if m is not None:
                value = m.group(1) + value[m.span(0)[1]:]
        value = datetime.datetime.fromisoformat(value)
    if value.tzinfo is None:
        return value.replace(tzinfo=datetime.timezone.utc)
    return value


def serialize_datetime(value: typing.Optional[datetime.datetime]) -> typing.Optional[str]:
    """
    Serialize datetime objects to string (with iso format).
    """
    if value is None:
        return None
    else:
        return value.isoformat(sep=' ', timespec='seconds')


datetime_with_tz = typing.Annotated[datetime.datetime,
                                    pydantic.BeforeValidator(force_utc),
                                    pydantic.PlainSerializer(serialize_datetime, return_type=typing.Optional[str])]


def force_utc_tuple(value: tuple) -> tuple:
    """
    Same as force_utc but for all elements of a tuple.
    """
    if value is None:
        return (None, None)
    r = []
    for i in range(len(value)):
        r.append(force_utc(value[i]))
    return tuple(value)


def serialize_datetime_tuple(value) -> typing.Optional[typing.List[typing.Optional[str]]]:
    """
    """
    if value is None:
        return None
    r = []
    for i in range(len(value)):
        r.append(serialize_datetime(value[i]))
    return r


datetime_tuple_with_tz = typing.Annotated[typing.Tuple[typing.Optional[datetime.datetime],
                                                       typing.Optional[datetime.datetime]],
                                          pydantic.BeforeValidator(force_utc_tuple),
                                          pydantic.PlainSerializer(
                                              serialize_datetime_tuple,
                                              return_type=typing.Optional[typing.List[typing.Optional[str]]])]


def get_dataframe_checker(_mode='Layer', **kwargs):
    """
    Checker for pandas DataFrame to be put in a ``data`` field.

    :param _mode: Point or Layer or Spectral
    :param kwargs: dict:
        keys : list of columns ot be accepted
        values : dict with constraints on the content of the columns, keys are:
            - translate: dict of replacement for values.
            - type: data type (float, int, str, etc.)
            - optional (bool): Optional or not (default is False)
            - min: for numeric types, the minimum value possible (included)
            - max: for numeric types, the maximim value possible (included)
            - nan_allowed: for numeric types, allow nan or not (default is False)
            - values: the list of accepted values
    """
    def check_dataframe(value, cls=None):
        # Check type -> ensure we have a pandas DataFrame
        if isinstance(value, dict):
            value = pd.DataFrame(value)
        elif isinstance(value, pd.DataFrame):
            value = value.copy()
        else:
            raise ValueError('data key should be a pandas DataFrame or a python dictionnary.')

        # Check columns
        columns = set(value.columns)
        if _mode == 'Layer':
            two_of_three = set(['top_height', 'bottom_height', 'thickness'])
            if len(columns.intersection(two_of_three)) == 3:
                if not (value['top_height'] - value['thickness'] == value['bottom_height']).all():
                    raise ValueError('Provided top_height, bottom_height and thickness that are inconsistent.')
            elif len(columns.intersection(two_of_three)) != 2 and columns != ('top_height'):
                raise ValueError(f'Should have 2 of three in {", ".join(two_of_three)}.')
            accepted_columns_min = set([])
            accepted_columns_max = set(['top_height', 'bottom_height', 'thickness'])
        elif _mode == "Spectral":
            accepted_columns_min = set(['min_wavelength', 'max_wavelength'])
            accepted_columns_max = accepted_columns_min
        elif _mode == 'Point':
            accepted_columns_min = set(['height'])
            accepted_columns_max = set(['height'])
        elif _mode == 'None':
            accepted_columns_min = set()
            accepted_columns_max = set()
        else:
            raise ValueError(f'Mode {_mode} unknown. The data model is ill-defined.')

        columns_min = []
        for k, v in kwargs.items():
            if 'optional' in v and v['optional']:
                continue
            else:
                columns_min.append(k)
        columns_min = set(columns_min) | accepted_columns_min
        columns_max = set(kwargs.keys()) | accepted_columns_max
        if not columns.issuperset(columns_min):
            raise ValueError(f'The data should contain at least the following columns: {", ".join(columns_min)}.')
        if not columns.issubset(columns_max):
            raise ValueError(f'The data should contain at most the following columns: {", ".join(columns_max)}.')

        # Depths processing
        # - Ensure types
        if _mode == 'Layer':
            height_keys = ['top_height', 'bottom_height', 'thickness']
        elif _mode == "Point":
            height_keys = ['height']
        elif _mode == "Spectral":
            height_keys = ['min_wavelength', 'max_wavelength']
        else:
            height_keys = []
        for key in height_keys:
            if key in columns:
                value[key] = value[key].astype('float')
        # - Completion of columns to ensure that top_height, bottom_height an dthickess are defined and coherent
        if _mode == 'Layer':
            # TODO: Reconstruct thickness from top_depth or bottom_depth if there is nan inside
            # or if only top_height is provided
            # (thickness is optional in a CAAML file)
            if 'top_height' not in columns:
                value['top_height'] = value['bottom_height'] + value['thickness']
            if 'bottom_height' not in columns:
                value['bottom_height'] = value['top_height'] - value['thickness']
            if 'thickness' not in columns:
                value['thickness'] = value['top_height'] - value['bottom_height']
        # - Ensure reasonnable values and no nan
        for key in height_keys:
            if pd.isna(value[key]).any():
                raise ValueError(f'Nan values are not allowed in {key} field')
            # For CAAML format we need to accept negative height values
            # if value[key].min() < 0:
            #     raise ValueError(f'Negative values for {key} is not accepted.')
            if _mode in ['Point', 'Layer'] and value[key].max() > 10:
                logging.warning(f'Values above 10m for {key}. Please check your data !')

        # Check other data
        for key, d in kwargs.items():
            if key not in value.columns:
                continue
            # Replace values if needed
            if 'translate' in d:
                value[key] = value[key].replace(d['translate'])
            # Check type
            _type = d['type'] if 'type' in d else 'float'
            value[key] = value[key].astype(_type)
            # Check min/max and nan presence for numeric types
            if np.issubdtype(value[key].dtype, np.number):
                # Check min/max
                if pd.isna(value[key].min()):
                    logging.warning(f'Data from key {key} is empty !')
                if 'min' in d:
                    _min = d['min']
                    if not pd.isna(value[key].min()) and value[key].min() < _min:
                        raise ValueError(f'Data from key {key} has unaccepted values (below {_min}).')
                if 'max' in d:
                    _max = d['max']
                    if not pd.isna(value[key].max()) and value[key].max() > _max:
                        raise ValueError(f'Data from key {key} has unaccepted values (above {_max}).')
                # Check nan presence
                nan_allowed = d['nan_allowed'] if 'nan_allowed' in d else False
                if not nan_allowed and pd.isna(value[key]).any():
                    raise ValueError(f'Nan values are not allowed in {key} field')
            # Check fixed allowed values if needed
            if 'values' in d:
                if not set(value[key].values).issubset(set(d['values'])):
                    raise ValueError(f'Unauthorized value for key {key}')

        if len(height_keys) > 0:
            value = value.sort_values(height_keys[0], ascending=False)

        return value

    return check_dataframe


class BaseData:
    """
    Base for classes with a data attribute of pandas DataFrame type.

    Manage the necessary functions to read, edit and dump/reload
    the data stored in the pandas DataFrame.
    """
    _data = typing.Optional[pd.DataFrame]

    @property
    def data(self) -> typing.Optional[pd.DataFrame]:
        """
        The profile data in the form of a Pandas Dataframe
        """
        if self._data is not None:
            return self._data.copy()
        else:
            return None

    @data.setter
    def data(self, value):
        checker = get_dataframe_checker(**self._data_config)
        self._data = checker(value)

    @data.deleter
    def data(self):
        self._data = None

    @pydantic.computed_field(alias='data', repr=True)
    @property
    def data_dict(self) -> typing.Optional[dict]:
        """
        The data in the form of a dictionnary.

        Useful for instance for JSON serialization
        """
        if self._data is None:
            return None
        return self._data.to_dict('list')

    @data_dict.setter
    def data_dict(self, value):
        self.data = value

    @data_dict.deleter
    def data_dict(self):
        del self.data


class BaseProfile(pydantic.BaseModel, BaseData):
    """
    Base class used for all profiles (stratigraphy, density, etc.)

    See the child class BaseProfile2 for all profiles except stratigraphy.
    """
    model_config = pydantic.ConfigDict(
        validate_assignment=True,
        extra='forbid',
        arbitrary_types_allowed=True)

    id: typing.Optional[str] = pydantic.Field(
        None,
        description="Unique identifier of the profile [A-Za-z0-9-]")
    name: typing.Optional[str] = pydantic.Field(
        None,
        description="Name/short description of the profile")
    related_profiles: typing.List[str] = pydantic.Field(
        [],
        description="id of related profiles")
    comment: typing.Optional[str] = pydantic.Field(
        None,
        description="General comment associated to the profile")
    record_time: typing.Optional[datetime_with_tz] = pydantic.Field(
        None,
        description="Time at which the profile was done (python datetime object).")
    record_period: datetime_tuple_with_tz = pydantic.Field(
        (None, None),
        description="Time period during which the profile was done "
        "(tuple of two python datetime object representing the begin time and end time).")
    profile_depth: typing.Optional[float] = pydantic.Field(
        None, ge=0,
        description="Total snow depth at the profile location, "
        "only if different from the general total snow depth reported in the metadata (m)")
    profile_swe: typing.Optional[float] = pydantic.Field(
        None, ge=0,
        description="SWE at the profile location, "
        "only if specific measurement at the precise location of the profile (mm or kg/m2)")
    additional_data: typing.Optional[AdditionalData] = pydantic.Field(
        None,
        description="Field to store additional data for CAAML compatibility (customData), do not use.")

    def __init__(self, data=None, data_dict=None, **kwargs):
        super().__init__(**kwargs)
        checker = get_dataframe_checker(**self._data_config)
        if data is not None:
            self._data = checker(data)
        elif data_dict is not None:
            self._data = checker(data_dict)
        else:
            raise ValueError('data key is required')


class BaseProfile2(BaseProfile):
    """
    Base class for all profiles except stratigraphy
    """
    quality_of_measurement: typing.Optional[typing.Literal[
        'Good', 'Uncertain', 'Low', 'Bad']] = pydantic.Field(
            None,
            description="Quality flag of the entire profile. See :ref:`uncertainty` for details.")
    uncertainty_of_measurement: typing.Optional[float] = pydantic.Field(
        None,
        gt = 0,
        description="Quantitative uncertainty of the entire profile (same units as ``data``). "
        "See :ref:`uncertainty` for details.")
    profile_nr: typing.Optional[int] = pydantic.Field(
        None, ge=0,
        description="Profile number (the lower is the higher priority)")


class BaseMergeable:
    """
    Implement a merge method for the pydantic data class.
    """
    def merge(self, other) -> None:
        """
        Function to merge with an object of the same type.

        Raise warnings when inconsistent data are encountered.

        Does not create a copy of the object.

        :param other: Other object of the same type
        :returns: None (merge done in-place)
        """
        # Loop on attributes
        for attr_name, attr_specs in type(self).model_fields.items():
            self._merge_attr(other, attr_name)

        # Special case of the _data key
        if hasattr(self, '_data'):
            self._merge_attr(other, '_data')

    def _merge_attr(self, other, attr_name):
        v_self = getattr(self, attr_name)
        v_other = getattr(other, attr_name)
        # 0. If all None, skip
        if v_self is None and v_other is None:
            return

        # 0. If other is None, keep the value of the reference (self)
        if v_other is None:
            return

        # 1. If one is None, use the other value
        if v_self is None and v_other is not None:
            setattr(self, attr_name, v_other)
            return

        # 2a. Special case: concatenate comments
        if 'comment' in attr_name:
            if isinstance(v_self, str) and isinstance(v_other, str):
                if v_self == v_other:
                    return

                v_new = v_self + '\n' + v_other
                setattr(self, attr_name, v_new)

        # 2b. Special case: _data key
        if attr_name == '_data' and isinstance(v_self, pd.DataFrame) and isinstance(v_other, pd.DataFrame):
            if v_self.equals(v_other):
                return
            elif len(v_other) == 0:
                return
            elif len(v_self) == 0:
                setattr(self, attr_name, v_other)
            else:
                logging.warning(f'Inconsistent data during merge between {self.__class__.__name__}.{attr_name}. '
                                'Possible data loss (not merged).')
            return

        # 3. Treatements based on type.
        # Raise warning if values are different and are merged with loss of data
        if isinstance(v_self, str) \
                or isinstance(v_self, int) or isinstance(v_self, float) \
                or isinstance(v_self, tuple) \
                or isinstance(v_self, datetime.datetime):
            if v_self == v_other:
                return  # This equality test could not be done on all data types, be careful.
            else:
                logging.warning(f'Inconsistent data during merge between {self.__class__.__name__}.{attr_name}. '
                                f'Values differ: {v_self} (reference) != {v_other} (merged)')
                return
        elif isinstance(v_self, list) and isinstance(v_other, list):
            v_new = v_self + v_other
            setattr(self, attr_name, v_new)
        elif hasattr(v_self, 'merge') and callable(v_self.merge):
            v_self.merge(v_other)
        else:
            logging.warning(f'Merge: Could not compare values of {self.__class__.__name__}.{attr_name}. '
                            'Possible data loss (not merged).')
            return



# -*- coding: utf-8 -*-

import typing

import pydantic

from snowprofile.classes import Time, Observer, Location, Weather, SurfaceConditions, Environment
from snowprofile.profiles import Stratigraphy, TemperatureProfile, DensityProfile, LWCProfile, \
    SSAProfile, SSAPointProfile, HardnessProfile, HardnessPointProfile, \
    StrengthProfile, ImpurityProfile, ScalarProfile, VectorialProfile
from snowprofile.stability_tests import CTStabilityTest, ECTStabilityTest, RBStabilityTest, PSTStabilityTest, \
    ShearFrameStabilityTest
from snowprofile._base_classes import AdditionalData, BaseMergeable

__all__ = ['SnowProfile']


class SnowProfile(pydantic.BaseModel, BaseMergeable):

    """
    The base class for representing and handling a snow profile.

    Data content
    ^^^^^^^^^^^^

    General information on snow profile and snow pit
    ''''''''''''''''''''''''''''''''''''''''''''''''

    id
      A unique id to identify the snow profile (optional, str, only [a-zA-Z0-9-])

    comment
      General comment on the overall measurement (optional, str)

    profile_comment
      General comment on the snow profile (optional, str)

    time
      Time of observation (:py:class:`snowprofile.classes.Time` object)

    observer
      Observer information (:py:class:`snowprofile.classes.Observer` object)

    location
      Location information (:py:class:`snowprofile.classes.Location` object)

    weather
      Weather observations (:py:class:`snowprofile.classes.Weather` object)

    surface_conditions
      Surface conditions (penetration depths, surface features, wind surface features, etc.)
      (:py:class:`snowprofile.classes.SurfaceConditions` object)

    profile_depth
      Depth of the profile (m)

    profile_depth_std
      Standard deviation of the profile depth around the snow pit (in case of multiple measurements, m)

    profile_swe
      Total SWE (mm or kg/m2)

    profile_swe_std
      Standard deviation of the total SWE (in case of mutilple measurements, mm or kg/m2)

    new_snow_24_depth
      Depth of the new snow from the past 24h (m)

    new_snow_24_depth_std
      Standard deviation of the depth of new snow (in case of multiple measurements, m)

    new_snow_24_swe
      SWE of the new snow from the past 24h (mm or kg/m2)

    new_snow_24_swe_std
      Standard deviation of the SWE of the new snow from the past 24h (in case of mutilple measurements, mm or kg/m2)

    snow_transport
      Presence and type of snow transport
       - No snow transport
       - Modified saltation: snow transport that remains confined close to the ground
       - Drifting snow: transport up to 6ft/2m
       - Blowing snow: transport above 6ft/2m

    snow_transport_occurence_24: typing.Optional[float] = pydantic.Field(None, ge=0, le=100)

    Profiles data
    '''''''''''''

    stratigraphy_profile
      The stratigraphy profile (unique, :py:class:`snowprofile.profiles.Stratigraphy` object)

    temperature_profiles:
      Temperature profiles (list of :py:class:`snowprofile.profiles.TemperatureProfile` objects)

    density_profiles:
      Density profiles (list of :py:class:`snowprofile.profiles.DensityProfile` objects)

    lwc_profiles:
      LWC profiles (list of :py:class:`snowprofile.profiles.LWCProfile` objects)

    ssa_profiles:
      SSA profiles (list of :py:class:`snowprofile.profiles.SSAProfile` or
      :py:class:`snowprofile.profiles.SSAPointProfile` objects)

    hardness_profiles:
      Hardness profiles (list of :py:class:`snowprofile.profiles.HardnessProfile` or
      :py:class:`snowprofile.profiles.HardnessPointProfile` objects)

    strength_profiles:
      Strength profiles (list of :py:class:`snowprofile.profiles.StrengthProfile` objects)

    impurity_profiles:
      Impurity profiles (list of :py:class:`snowprofile.profiles.ImpurityProfile` objects)

    stability_tests: List of stabilty tests. See :ref:`Stability tests` for details.

    Other data
    ''''''''''

    application
      Information on the application or code that generated the profile (optional, str)

    application_version
      Version of the application or code that generated the profile (optional, str)

    profiles_comment
      Comment associated to profiles, for CAAML compatibility only, do not use (str)

    additional_data and profile_additional_data
      Room to store additional data for CAAML compatibility (customData), do not use.

    """

    model_config = pydantic.ConfigDict(
        validate_assignment=True,
        extra='forbid')

    id: typing.Optional[str] = None
    comment: typing.Optional[str] = None
    profile_comment: typing.Optional[str] = None
    time: Time = Time()
    observer: Observer = Observer()
    location: Location = Location(name='Unknown')
    environment: Environment = Environment()
    application: typing.Optional[str] = 'snowprofile'
    application_version: typing.Optional[str] = None
    profile_depth: typing.Optional[float] = pydantic.Field(None, ge=0)
    profile_depth_std: typing.Optional[float] = pydantic.Field(None, ge=0)
    profile_swe: typing.Optional[float] = pydantic.Field(None, ge=0)
    profile_swe_std: typing.Optional[float] = pydantic.Field(None, ge=0)
    new_snow_24_depth: typing.Optional[float] = pydantic.Field(None, ge=0)
    new_snow_24_depth_std: typing.Optional[float] = pydantic.Field(None, ge=0)
    new_snow_24_swe: typing.Optional[float] = pydantic.Field(None, ge=0)
    new_snow_24_swe_std: typing.Optional[float] = pydantic.Field(None, ge=0)
    snow_transport: typing.Optional[typing.Literal[
        'No snow transport',
        'Modified saltation',
        'Drifting snow',
        'Blowing snow']] = None
    snow_transport_occurence_24: typing.Optional[float] = pydantic.Field(None, ge=0, le=100)
    weather: Weather = Weather()
    surface_conditions: SurfaceConditions = SurfaceConditions()
    stratigraphy_profile: typing.Optional[Stratigraphy] = None
    temperature_profiles: typing.List[TemperatureProfile] = []
    density_profiles: typing.List[DensityProfile] = []
    lwc_profiles: typing.List[LWCProfile] = []
    ssa_profiles: typing.List[SSAProfile | SSAPointProfile] = []
    hardness_profiles: typing.List[HardnessProfile | HardnessPointProfile] = []
    strength_profiles: typing.List[StrengthProfile] = []
    impurity_profiles: typing.List[ImpurityProfile] = []
    other_scalar_profiles: typing.List[ScalarProfile] = []
    other_vectorial_profiles: typing.List[VectorialProfile] = []
    stability_tests: typing.List[
        CTStabilityTest | ECTStabilityTest | RBStabilityTest | PSTStabilityTest | ShearFrameStabilityTest] = []
    additional_data: typing.Optional[AdditionalData] = None
    profile_additional_data: typing.Optional[AdditionalData] = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import os.path
import unittest
import tempfile

import snowprofile

_here = os.path.dirname(os.path.realpath(__file__))


def gen_snowprofile():
    sp = snowprofile.SnowProfile()
    sp.location.name = 'Col de Porte'
    sp.location.latitude = 45.295043
    sp.location.longitude = 5.76559
    import datetime
    sp.time.record_time = datetime.datetime(2019, 12, 25, 10, 0)
    sp.time.report_time = datetime.datetime.now()

    dp = snowprofile.profiles.DensityProfile(
        method_of_measurement="Snow Cutter",
        quality_of_measurement="Good",
        probed_thickness=0.03,  # 3cm cutter thickness
        data = {'top_height': [1.2, 1.1, 1.0, 0.50],
                'thickness': [0.1, 0.1, 0.5, 0.5],
                'density': [75, 100, 180, 230]}
    )
    sp.density_profiles.append(dp)
    return sp


class TestIOInternalJSON(unittest.TestCase):

    def test_to_from_snowprofile(self):
        sp = gen_snowprofile()

        json = snowprofile.io.to_json(sp)
        spn = snowprofile.io.from_json(json)

        assert spn.location.name == sp.location.name, 'Names differ'
        assert spn.time.record_time == sp.time.record_time, 'Time differs'
        assert sp.density_profiles[0].data.equals(spn.density_profiles[0].data), 'Density profiles differ'

    def test_write_read_internal_json(self):
        sp = gen_snowprofile()

        fd = tempfile.mkstemp(prefix='snowprofile-tests', suffix='.json')
        fn = snowprofile.io.write_internal_json(sp, fd[1])
        spn = snowprofile.io.read_internal_json(fn)

        assert spn.location.name == sp.location.name, f'Names differ, see {fn}'
        assert spn.time.record_time == sp.time.record_time, f'Time differs, see {fn}'
        assert sp.density_profiles[0].data.equals(spn.density_profiles[0].data), f'Density profiles differ, see {fn}'

        os.remove(fn)

    def test_to_from_real_data(self):
        sp = snowprofile.io.read_caaml6_xml(os.path.join(_here, 'resources', 'SnowProfile_IACS_SLF22950.xml'))
        json = snowprofile.io.to_json(sp)
        spn = snowprofile.io.from_json(json)

        assert len(sp.stability_tests) == len(spn.stability_tests)
        for i in range(len(sp.stability_tests)):
            assert isinstance(spn.stability_tests[i], type(sp.stability_tests[i]))


if __name__ == "__main__":
    unittest.main()
